/* Just another XML Processor
Copyright (C) 2006-08 Benito van der Zander

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
import java.io.*;
import java.util.*;
import java.text.*;
//import javax.xml.parsers.SAXParserFactory;
//import javax.xml.parsers.SAXParser;
//import javax.xml.parsers.ParserConfigurationException;
//import org.xml.sax.XMLReader;
//import org.xml.sax.SAXException; 
//import org.xml.sax.helpers.DefaultHandler;
//import org.xml.sax.Attributes;
import org.w3c.dom.*;

//import com.sun.org.apache.xerces.internal.impl.msg.XMLMessageFormatter;

//import sun.security.krb5.internal.crypto.b;

//import com.sun.org.apache.xpath.internal.compiler.FunctionTable;

//import com.sun.org.apache.xerces.internal.xni.XMLLocator;

//import sun.security.krb5.internal.r;

import javax.xml.parsers.*;

//==============FileParser implementation==============
  
public class FileParser{
  final static int MAX_FOR_EACH_LENGTH=8192;
  final static boolean TALK=true;
  /*final static */boolean VERBOSE=false;
  final static boolean VERBOSE_PARSE_STRING=!true;
  final static boolean VERBOSE_FOREACH=!true;
  final static boolean VERBOSE_VARIABLE_CHANGES=!true;
  final static boolean VERBOSE_DEBUG_PARSER=!true;
  boolean printChangedOutputFiles = false;
  
  private DocumentBuilderFactory factory;
  private DocumentBuilder        builder;

        
      
  private HashMap<String,Variable> variables=new HashMap<String,Variable>();
  private HashMap<String,String> preferList=new HashMap<String,String>();
  private HashMap<String,FunctionDefinitionCmd> functions=new HashMap<String,FunctionDefinitionCmd>();
  //HashMap<String, CommandList> fileFunctions = new HashMap<String,CommandList>();
  //private HashMap<String,File> myLoadedFiles=new HashMap<String,File>();; //my to different from parent
  
  private long codeLastModified; //last modified of files loaded before transforming (include transformation file date)
  private HashMap<String,Long> xmlVariableModified=new HashMap<String,Long>();
  private Properties fileDates=new Properties();
  private Properties oldFileDates=new Properties();
  private StringBuffer additionalLoadedFiles=new StringBuffer();
  
  long getLastModifiedXML(){
    long result=0;
    for (Long mod : xmlVariableModified.values()) 
      if (mod>result) result=mod;
    return result;
  }
  
  long getRealLastModified(){
    long files=getLastModifiedXML();
    return (files>codeLastModified)?files:codeLastModified;
  }
  
  String formatDate(long date){
    return (new SimpleDateFormat("yyyy-MM-dd HH:mm:ssZ")).format(date);
  }
  
  String readVariable(String v, boolean mustExist) throws Exception{ //Ermittelt den Wert einer Variablen
    String[] parts = v.split("/");
    if ((parts.length==2) && (parts[1].equals("$lastModified"))) 
      if (xmlVariableModified.containsKey(parts[0])) 
        return formatDate(xmlVariableModified.get(parts[0]));
        
    if (variables.get(v)!=null && variables.get(v).string!=null) return variables.get(v).string;
    
    Element el = variables.get(parts[0]).element;
      
    for (int i=1;i<parts.length;i++) {
      if (el==null) return mustExist?null:""; 
      if (parts[i].charAt(0) == '@') {
        if (i!=parts.length-1) throw new Exception("Attribut muss letzter Teil in Variable "+v+" sein.");
        return el.getAttribute(parts[i].substring(1));
      } else {
        NodeList nl=el.getChildNodes();
        if (nl==null) return mustExist?null:""; 
        el=bestElement(nl,parts[i]);
      }
    } 
    if ((el==null)||(el.getFirstChild()==null)) return mustExist?null:"";
    return el.getFirstChild().getNodeValue(); 
  }
  String callFunction(FunctionDefinitionCmd func, ArrayList<Parameter> params) throws Exception{
    if (func==null) return "";
    HashMap<String,Variable> varsave=new HashMap<String, Variable>();
    varsave.putAll(variables);//  new HashMap<String,Element>();
    HashMap<String,Long> varmodsave=new HashMap<String, Long>();
    varmodsave.putAll(xmlVariableModified);//  new HashMap<String,Element>();
    if (params!=null){
      if (params.size()>func.params.size())
        throw new EParseException("to many parameters ("+params.size()+") for "+func.name+" ("+func.params.size()+")");
      for (int i=0;i < params.size();i++) 
        variables.put(func.params.get(i).raw(), variables.get(params.get(i).raw()));
    }
    //HashMap<String, Element> varSave=(HashMap<String, Element>)(xmlVariables.clone());
    String res=func.body.performAll();
    variables=varsave;
    //if (!xmlVariableModified.equals(varmodsave)) {
      xmlVariableModified=varmodsave;
    
    return res;
  }
  Element bestElement(NodeList nl, String name){
    if (nl.getLength()<=0) return null;
    else if (nl.getLength()==1) {
      if (nl.item(0).getNodeName().equals(name)&&(nl.item(0).getNodeType()==Node.ELEMENT_NODE)) return (Element)nl.item(0);
    }else {
      Element e=null;
      int v=0;
      for (int i=0;i<nl.getLength();i++){
        if (!name.equals(nl.item(i).getNodeName())||(nl.item(i).getNodeType()!=Node.ELEMENT_NODE)) continue;
        int nv=1;
        Element ne=(Element)(nl.item(i));
        for (String a: preferList.keySet()) 
          if (ne.getAttribute(a).equals(preferList.get(a))) nv++;
        if (nv>v) {
          e=ne;
          v=nv;
        }
      }
      return e;
    }
    return null;
  }
  static enum StringPartType {STRING, VARIABLE};
  class StringPart{
    StringPartType type;
    String text;
    public StringPart(StringPartType type,String value) {
	    this.type=type;
	    this.text=value;
	  }
  }
  class Variable{
    Element element;
    String string;
    Variable(Element e){
      element=e;
    }
    Variable(String s){
      string=s;
    }
  }
  class Parameter {
    ArrayList<StringPart> strParts= new ArrayList<StringPart>();
    void parseStr(String s) throws EParseException{
      if (VERBOSE_PARSE_STRING) System.out.println("Current String: "+s);
      //you can mix string with variables to get a filename, if var == "abc", then this 
      // "c:\"var".txt"
      // is "c:\abc.txt"
      boolean isvar=s.charAt(0)!='"';
      StringBuffer curText=new StringBuffer();
      curText.ensureCapacity(s.length());
      for (int i=isvar?0:1;i<s.length();i++){
        char c = s.charAt(i);
        switch (c) {
          case '"': //change between variable/string
            strParts.add(new StringPart(isvar?StringPartType.VARIABLE:StringPartType.STRING,curText.toString()));
            if (VERBOSE_PARSE_STRING) System.out.println("Current part: "+(isvar?"var: "+curText:"string: \""+curText+"\""));
            curText.setLength(0);
            isvar = !isvar;
            break;
          case '\\': //escape character
            if (!isvar) { //only in strings
              if (i+1==s.length()) throw new EParseException("escape character \\ doesn't escape anything in string: "+s);
              i++;
              c=s.charAt(i);
              if (c=='"') curText.append('"');
              else if (c=='\\') curText.append('\\');
              else if (c=='\n') curText.append('\n');
              else if (c=='\t') curText.append('\t');
              else if (c=='\r') curText.append('\r');
              else throw new EParseException("escaped character \"\\"+s.charAt(i)+"\" is unknown in string: "+s);
              break;
            } else ; //fall through
          default: //just add character
            curText.append(c);
        }
      }
      if (curText.length()>0){
        if (!isvar) throw new EParseException("missing closing \" in string: "+s);
        strParts.add(new StringPart(StringPartType.VARIABLE,curText.toString()));
      }
    }
    public Parameter(String s, boolean parse) throws EParseException{
      if (!parse) strParts.add(new StringPart(StringPartType.STRING,s));
      else parseStr(s);
    }
    public Parameter(String s) throws EParseException{
      parseStr(s);
    }
    String raw(){
      assert(strParts.size()==1);
      return strParts.get(0).text;
    }
    String value() throws Exception{ //Ermittelt den Wert der Variablenteile
      StringBuffer result=new StringBuffer();
      for (StringPart p: strParts)
        if (p.type==StringPartType.STRING) result.append(p.text);
        else {
          if (p.text.equals("$lastModified")) 
            return formatDate(getRealLastModified()); 
          else if (functions.containsKey(p.text)) result.append(callFunction(functions.get(p.text),null));
          else if (variables.containsKey(p.text.split("/")[0])) result.append(readVariable(p.text,false));
          else {
            StringBuffer vars=new StringBuffer();
            for (String varname : variables.keySet()) {
              vars.append(varname+" ");
            }
            throw new EParseException("Unknown variable/function '"+p.text+"' (check variable '"+p.text.split("/")[0]+"')\n"+"Existing variables: "+vars);
            
          }
        }
      return result.toString();
    }
  }
  abstract class Command{
    ArrayList<Parameter> params=new ArrayList<Parameter>();
    String file;
    int line;
    abstract String perform() throws Exception;
  }
  class LoadXMLCmd extends Command{
    String perform()throws Exception
    {
      if (!":".equals(params.get(1).raw()))
        throw new EParseException("Expect ':' in "+file+":"+line+", found "+params.get(1).raw());
      
      String varname = params.get(0).raw();
      String fileName = params.get(2).value();
      //System.out.println(fileName);
      Document document;
      try {
        document = builder.parse(fileName);
      } catch (FileNotFoundException e) {
        throw new EParseException("Can't open XML file: "+fileName+"\n"+
                                  file+":"+line);
      }
      
      Element root = document.getDocumentElement();
      variables.put(varname,new Variable(root));
      xmlVariableModified.put(varname,(new File(fileName)).lastModified());
      
      additionalLoadedFiles.append(fileName+"?");

      //myLoadedFiles.put(fileName,new File(fileName));
      return "";
      //if (VERBOSE) System.out.println("Load "+filename+" as "+varname);
    };
  }
  class WriteCmd extends Command{
    WriteCmd(){}
    WriteCmd(Parameter ps){
      params.add(ps);
    }
    String perform() throws Exception{
      try {
        return params.get(0).value();
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $WRITE in "+file+":"+line, e);
      }
    }
  }
  class WriteXCmd extends Command{
    WriteXCmd(){}
    WriteXCmd(Parameter ps){
      params.add(ps);
    }
    String perform() throws Exception{
      try {
        if (params.get(0).value().indexOf("{{$")<0) return params.get(0).value(); //no need to parse
        else return parse(new StringReader(params.get(0).value())).body.performAll(); //parse and execute
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $WRITE in "+file+":"+line, e);
      }
    }
  }
  class PreferCmd extends Command{
    String perform() throws Exception {
      try {
        if (!"==".equals(params.get(1).raw()))
          throw new EParseException("Expect '==', but found "+params.get(1).raw());
        preferList.put(params.get(0).value(),params.get(2).value());
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $PREFER in "+file+":"+line, e);
      }
      return "";
    }
  }
  class ImportCmd extends Command{
    String perform()throws Exception{
      try {
        if (!":".equals(params.get(1).raw()))
          throw new EParseException("Expect ':', but found "+params.get(1).raw());
        String fileIn=params.get(2).value();

        FunctionDefinitionCmd temp=parseFile(fileIn);
        functions.put(params.get(0).raw(), temp);
  
        additionalLoadedFiles.append(fileIn+"?");
        
        //if (TALK||VERBOSE) System.out.println("Transformed to "+fileOut);
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $IMPORT in "+file+":"+line, e);
      }

      return "";
    }
  }
  class OutputCmd extends Command{
    String perform() throws Exception{      
      try {
        String fileOut=params.get(1).value();
        StringBuffer oldLoadedFiles=additionalLoadedFiles;
        additionalLoadedFiles=new StringBuffer();
        long currentFileDate= getRealLastModified();
        long lastFileDate;
        if (oldFileDates.getProperty("$OUTPUT:NEED:"+fileOut)!=null) {
          String[] needs=oldFileDates.getProperty("$OUTPUT:NEED:"+fileOut).split("\\?");
          for (String string : needs) {
            if (!string.equals("")) 
              if (currentFileDate<(new File(string)).lastModified())
                currentFileDate=(new File(string)).lastModified();
          }
        }
        if (oldFileDates.getProperty("$OUTPUT:"+fileOut)!=null)
          lastFileDate=Long.parseLong((oldFileDates.getProperty("$OUTPUT:"+fileOut)));
        else
          lastFileDate=0;
        if (currentFileDate<lastFileDate) {
          File out = new File(fileOut);
          if (out.exists() && (out.lastModified()<=lastFileDate)){ 
            if (VERBOSE) System.out.println("output to "+fileOut+" skipped (no change)");
            fileDates.put("$OUTPUT:"+fileOut, (new File(fileOut)).lastModified()+"");
            if (oldFileDates.getProperty("$OUTPUT:NEED:"+fileOut)!=null)
              fileDates.put("$OUTPUT:NEED:"+fileOut, 
                            oldFileDates.getProperty("$OUTPUT:NEED:"+fileOut));
            additionalLoadedFiles=oldLoadedFiles;
            return "";
          }
        }
           
        FileOutputStream fos = new FileOutputStream(fileOut);
        OutputStreamWriter fileStream ;
        if (params.size()==2) 
          fileStream = new OutputStreamWriter(fos); 
         else
          fileStream = new OutputStreamWriter(fos,params.get(2).value()); 
        fileStream.write(params.get(0).value());
        fileStream.close();
        fos.close();

        if (VERBOSE) System.out.println("completed output to "+fileOut);
        if (printChangedOutputFiles) System.out.println(fileOut);

        fileDates.put("$OUTPUT:"+fileOut, (new File(fileOut)).lastModified()+"");
        if (additionalLoadedFiles.length()>0)
          fileDates.put("$OUTPUT:NEED:"+fileOut, additionalLoadedFiles.toString());
        
        additionalLoadedFiles=oldLoadedFiles;
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $OUTPUT in "+file+":"+line, e);
      }
      //transformationProperties.setProperty(fileIn+" to "+fileOut,(new Long(out.lastModified())).toString());
      
      return "";
    }
  }
  class ForeachCmd extends Command{
    CommandList loopBody=new CommandList();
    String perform() throws Exception{
      try {
        String var=params.get(0).raw();
        if (!":".equals(params.get(1).raw()))
          throw new EParseException("Expect ':', but found "+params.get(1).raw());
        //if (xmlVariables.containsKey(var)) 
          //throw new EParseException("foreach variable "+var+" is already set");
        String variablePath=params.get(2).raw();
        
        StringTokenizer st = new StringTokenizer(variablePath,"/");
        Element el = variables.get(st.nextToken()).element;
  
        NodeList nl=null;
        if (el!=null) {
          String elName="";
          while (st.hasMoreTokens()) {
            if (el==null) break; 
            nl=el.getChildNodes();
            if (nl==null) break; 
            elName=st.nextToken();
            el=bestElement(nl,elName);
            //if (VERBOSE_FOREACH) System.out.println("   Search next "+((el!=null)?el.getNodeName():"<null>"));
          }
          if (nl!=null) {
            StringBuffer result=new StringBuffer();
            for (int i=0;i<nl.getLength();i++) 
              if ((nl.item(i).getNodeType()==Node.ELEMENT_NODE)&&elName.equals(nl.item(i).getNodeName())) {
                variables.put(var,new Variable((Element)nl.item(i)));
                if (VERBOSE_VARIABLE_CHANGES)
                  System.out.println("foreach sets "+var+" "+file+":"+line);
                result.append(loopBody.performAll());
              }
            variables.remove(var);
            if (VERBOSE_VARIABLE_CHANGES)
              System.out.println("foreach removes "+var+" "+file+":"+line);
            return result.toString();
          }
        }
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $FOREACH in "+file+":"+line,e);
      }
      return "";
    }
  }
  static enum BooleanEvaluation{BE_TRUE, BE_FALSE, BE_OR, BE_AND, BE_OPEN, BE_CLOSE};
  class IfElseCmd extends Command{
    CommandList ifBody=new CommandList();
    CommandList elseBody=new CommandList();
    boolean _elseBodyRead=false;
    BooleanEvaluation evalBoolean(BooleanEvaluation v1, BooleanEvaluation op, BooleanEvaluation v2/*rock*/) throws EParseException{
      if (op == BooleanEvaluation.BE_OR)
         return ((v1 == BooleanEvaluation.BE_TRUE) || (v2 == BooleanEvaluation.BE_TRUE))?BooleanEvaluation.BE_TRUE:BooleanEvaluation.BE_FALSE;
        else if (op == BooleanEvaluation.BE_AND)
         return ((v1 == BooleanEvaluation.BE_TRUE) && (v2 == BooleanEvaluation.BE_TRUE))?BooleanEvaluation.BE_TRUE:BooleanEvaluation.BE_FALSE;
        else
         throw new EParseException("impossible");
    }
    String perform() throws Exception{
      try {
        //evaluate if condition
        //Grammar: TERM = VAL BIN-OP VAL | TERM && TERM | TERM || TERM | (TERM)
        //It uses a simple stack approach, everything is pushed on the stack and
        //evaluated asap (if closed bracket or complete bin op occurs).
        //Values (strings) are however not pushed on the stack, but trigger a
        //look-ahead to the bin-op and then the result of the binop is pushed
        LinkedList<BooleanEvaluation> stack = new LinkedList<BooleanEvaluation>();
	for (int i=0;i<params.size();i++){
          String cur = params.get(i).raw();
          if ("(".equals(cur)) stack.addFirst(BooleanEvaluation.BE_OPEN);
          else if (")".equals(cur)) {
	      BooleanEvaluation oldValue1 = stack.pollFirst();
	      if (oldValue1 == null || oldValue1 != BooleanEvaluation.BE_TRUE && oldValue1 != BooleanEvaluation.BE_FALSE)
	        throw new EParseException("Invalid expressions in boolean formula (1). around "+params.get(i).raw());
	      BooleanEvaluation oldOp = stack.pollFirst();
	      if (oldOp == BooleanEvaluation.BE_OPEN) {
		stack.addFirst(oldValue1);
	      } else {
                if (oldOp == null || oldOp != BooleanEvaluation.BE_AND && oldOp != BooleanEvaluation.BE_OR)
                  throw new EParseException("Invalid expressions in boolean formula (2). around "+params.get(i).raw());
                BooleanEvaluation oldValue2 = stack.pollFirst();
                if (oldValue2 == null || oldValue2 != BooleanEvaluation.BE_TRUE && oldValue2 != BooleanEvaluation.BE_FALSE)
                  throw new EParseException("Invalid expressions in boolean formula (3). around "+params.get(i).raw());
                if (stack.pollFirst() != BooleanEvaluation.BE_OPEN)
                  throw new EParseException("Invalid expressions in boolean formula (4). around "+params.get(i).raw());

                stack.addFirst(evalBoolean(oldValue1, oldOp, oldValue2));
              }
          } else if ("||".equals(cur)) stack.addFirst(BooleanEvaluation.BE_OR);
          else if ("&&".equals(cur)) stack.addFirst(BooleanEvaluation.BE_AND);
          else {
            boolean newValue;
            if ("==".equals(params.get(i+1).raw())) {
              newValue=(params.get(i+0).value().equals(params.get(i+2).value()));
            } else if ("!=".equals(params.get(i+1).raw())) {
              newValue=(!params.get(i+0).value().equals(params.get(i+2).value()));
            } else
              throw new EParseException("Expected '==' or '!=', but found "+params.get(i+1).raw());
	    if (stack.isEmpty() || stack.getFirst() == BooleanEvaluation.BE_OPEN)
	      stack.addFirst(newValue?BooleanEvaluation.BE_TRUE:BooleanEvaluation.BE_FALSE);
	    else {
	      BooleanEvaluation oldOp = stack.pollFirst();
	      if (oldOp != BooleanEvaluation.BE_AND && oldOp != BooleanEvaluation.BE_OR)
	        throw new EParseException("Invalid expressions in boolean formula (5). around "+params.get(i).raw());
	      BooleanEvaluation oldValue = stack.pollFirst();
	      if (oldOp == null || oldValue != BooleanEvaluation.BE_TRUE && oldValue != BooleanEvaluation.BE_FALSE)
	        throw new EParseException("Invalid expressions in boolean formula (6). around "+params.get(i).raw());
	      if (oldOp == BooleanEvaluation.BE_OR) stack.addFirst((newValue || (oldValue == BooleanEvaluation.BE_TRUE))?BooleanEvaluation.BE_TRUE:BooleanEvaluation.BE_FALSE);
	      else if (oldOp == BooleanEvaluation.BE_AND) stack.addFirst((newValue && (oldValue == BooleanEvaluation.BE_TRUE))?BooleanEvaluation.BE_TRUE:BooleanEvaluation.BE_FALSE);
	      else throw new EParseException("impossible");
	    }
	    i+=2;
          }
        }
        if (stack.isEmpty()) throw new EParseException("empty if condition");
        BooleanEvaluation curVal = stack.pollFirst();
        if (curVal == BooleanEvaluation.BE_OPEN) throw new EParseException("unclosed bracket in if condition");
        if (curVal == BooleanEvaluation.BE_AND) throw new EParseException("&& without value in if condition");
        if (curVal == BooleanEvaluation.BE_OR) throw new EParseException("|| without value in if condition");
        while (!stack.isEmpty()) {
            BooleanEvaluation op = stack.pollFirst();
          if (op != BooleanEvaluation.BE_AND && op != BooleanEvaluation.BE_OR) throw new EParseException("invalid operator in if condition: "+op);
            BooleanEvaluation val = stack.pollFirst();
          if (val == null) throw new EParseException("missing value in if condition");
          if (val != BooleanEvaluation.BE_TRUE && val != BooleanEvaluation.BE_FALSE) throw new EParseException("invalid value in if condition: "+val);
          curVal = evalBoolean(curVal, op, val);
        }

        boolean result=curVal == BooleanEvaluation.BE_TRUE;
        if (result) return ifBody.performAll();
        else return elseBody.performAll();
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $IF in "+file+":"+line, e);
      }
        
    }
  }
  class CallCmd extends Command{
    String executeProcess(String prog, String stdinput) throws EParseException{
    //System.out.println("cmd: "+prog+ " stdin: "+stdinput);
      try {

        Process process = Runtime.getRuntime ().exec (prog);
        InputStream stdout = process.getInputStream ();

        OutputStream stdin = process.getOutputStream ();
        stdin.write(stdinput.getBytes() );
        stdin.close();
        
        StringBuffer out = new StringBuffer();
        byte[] b = new byte[4096];
        for (int n; (n = stdout.read(b)) != -1;) out.append(new String(b, 0, n));
        return out.toString();
      } catch (IOException e) {
        throw new EParseException("Couldn't execute command "+prog+" with input "+stdinput);
      }
      
    }
    String perform()throws Exception{
      try {
        StringBuffer cmd = new StringBuffer();
        String data = new String();
        StringTokenizer st = new StringTokenizer(params.get(0).value()+"|","\"|",true);
        boolean inStr=false;
        while (st.hasMoreTokens()) {
          String nextToken = st.nextToken();
          if (!inStr && "|".equals(nextToken)) {
            data=executeProcess(cmd.toString(),data);
            cmd.setLength(0);
          } else {
            if ("\"".equals(nextToken)) inStr=!inStr;
            cmd.append(nextToken);
          }
        }

        if (params.size()>=1) variables.put(params.get(1).raw(), new Variable(data));
        else return data;
        //if (TALK||VERBOSE) System.out.println("Transformed to "+fileOut);
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in $CALL in "+file+":"+line, e);
      }

      return "";
    }
  }
  class FunctionCmd extends Command{
    FunctionDefinitionCmd definition;
    public FunctionCmd(FunctionDefinitionCmd def) {
      definition=def;
    }
    String perform() throws Exception{
      try {
        return callFunction(definition, params);
      } catch (EParseException e) {
        throw new EParseException(e.getMessage()+"\n in function call of "+definition.name+" in "+file+":"+line, e);
      }
    }
  }
  class FunctionDefinitionCmd extends Command{
    CommandList body=new CommandList();
    String name;
    String perform(){return "";}
  }
  class CommandList extends ArrayList<Command>{
     private static final long serialVersionUID = -5410106954825197447L;

    String performAll() throws Exception{
      StringBuffer result=new StringBuffer();
      for (Command com: this) result.append(com.perform());
      return result.toString();
    }
  }
 
  class EParseException extends Exception{
    private static final long serialVersionUID = -7299044713974740222L;

    EParseException (String e){
      super(e);
    }
    EParseException (String e, EParseException previous){
      super(e,(VERBOSE && VERBOSE_DEBUG_PARSER)?previous:null);      
    }
  };

  FileParser()throws Exception{
    factory  = DocumentBuilderFactory.newInstance();
    builder  = factory.newDocumentBuilder();
  }  
        
  //---------------parsing---------------
  String fileName=""; //for error messages
  public FunctionDefinitionCmd parseFile(String fileName) throws Exception{
    if (VERBOSE) System.out.println("Started parsing file "+fileName);
    
    File parsedFile=new File(fileName);
    if (codeLastModified<parsedFile.lastModified()) 
      codeLastModified=parsedFile.lastModified();
    
    this.fileName=fileName;
    FunctionDefinitionCmd  temp = parse(new FileReader(fileName));
    this.fileName=null;
    return temp;
  }
  public FunctionDefinitionCmd parse(Reader reader) throws Exception{     
    BufferedReader buffer=new BufferedReader(reader);
    
    FunctionDefinitionCmd result=new FunctionDefinitionCmd();
    CommandList commands=result.body;
    Stack<Command> cmdStack=new Stack<Command>();
    cmdStack.add(result);
    
    boolean ignoreWhiteSpace=false;
    boolean ignoreAllWhiteSpace=false;
    //COMMAND command=CMD_NONE;
    StringBuffer read=new StringBuffer();
    int ch3;
    int ch2=buffer.read();
    int ch1=buffer.read();
    read.append((char)ch2);
    read.append((char)ch1);
    int line=1;
    while (ch1!=-1) {
      ch3=ch2;
      ch2=ch1;
      ch1=buffer.read();
      if (ch1==-1) break;
      //System.out.print((char)ch1);
      if (!ignoreWhiteSpace || ((ch1!=' ') && (ch1!='\t')&& (ch1!='\r') && (ch1!='\n'))) {
        read.append((char)ch1);
        if (ignoreWhiteSpace && !ignoreAllWhiteSpace) ignoreWhiteSpace=false;
      }
      if (ch1=='\n') line++;
      if ((ch3=='{') && (ch2=='{') && (ch1=='$')) { //new command read
        read.delete(read.length()-3, read.length());
        if (read.length()>0){ //text before as write command
          //System.out.println("implicit: "+read);
          commands.add(new WriteCmd(new Parameter(read.toString(),false)));
          read=new StringBuffer();
        }
        ignoreWhiteSpace=false;
        ArrayList<String> parts= new ArrayList<String>();
        int ch;
        boolean skipRead=false;
        ch=-1;
        //parse command name and parameters
        do{ 
          if (!skipRead) ch=buffer.read();
          while (ch==' ') ch=buffer.read();
          StringBuffer sb=new StringBuffer();
          boolean inString=false;
          while (inString||(ch=='_')||(ch=='/')||(ch=='-')||(ch=='"')||(ch=='$')||(ch=='@')||
                   ((ch>='0')&&(ch<='9'))||((ch>='a')&&(ch<='z'))||((ch>='A')&&(ch<='Z'))) {
            //System.out.print((char)(ch));
            sb.append((char)ch);
            if (inString && ch=='\\') {
              ch=buffer.read();
              if (ch!='"' && ch!='\\' && ch!='n' && ch!='r' && ch!='t')  
                throw new EParseException("unknown escape character \\"+ch+"\n in "+fileName+":"+line);
              sb.append((char)ch);
            } else {
              if (ch=='"') inString=!inString;
            }
            ch=buffer.read();
	    if (ch==-1)
	      throw new EParseException("string parameter didn't end (probably wrongly escaped parameter) in "+fileName+":"+line);
          }
          if (sb.length()>0)
            parts.add(sb.toString());
          
          sb=new StringBuffer();
	  boolean continueLoop = true;
          while (continueLoop && ((ch=='=')||(ch==':')||(ch=='!')||(ch=='|')||(ch=='&')||(ch=='(')||(ch==')'))){
            sb.append((char)ch);
	    continueLoop = (ch!='(')&&(ch!=')');
            ch=buffer.read();
            skipRead=true;
          }
          if (sb.length()>0)
            parts.add(sb.toString());
          
          if (ch=='\n') line++;
        } while (ch!='}' && ch!=-1);
        Command command=null;
        CommandList newcommands=null;
        String commandstr=parts.get(0);//.toUpperCase();
        //for (String str: parts) 
        //  System.out.print("`"+str+"`  ");
        //System.out.println();
        parts.remove(0);
        
        //recognize commands
        
        if (commandstr.equals("LOADXML")) command=new LoadXMLCmd();
        else if (commandstr.equals("IMPORT")) command=new ImportCmd();
        else if (commandstr.equals("OUTPUT")) command=new OutputCmd();
        else if (commandstr.equals("PREFER")) command=new PreferCmd();
        else if (commandstr.equals("WRITE")) command=new WriteCmd();
        else if (commandstr.equals("WRITEX")) command=new WriteXCmd();
        else if (commandstr.equals("CALL")) {
          command=new CallCmd(); 
        } else if (commandstr.equals("IF")) {
          command=new IfElseCmd();
          cmdStack.push(command);
          newcommands=((IfElseCmd)command).ifBody;
        } else if (commandstr.equals("DEFINE")) {
          FunctionDefinitionCmd def=new FunctionDefinitionCmd();
          command=def;
          def.name=parts.get(0);
          parts.remove(0);
          cmdStack.push(def);
          newcommands=def.body;
          functions.put(def.name,def);
        } else if (commandstr.equals("FOREACH")) {
          command=new ForeachCmd();
          cmdStack.push(command);
          newcommands=((ForeachCmd)command).loopBody;
        } else if (commandstr.equals("/FOREACH")||commandstr.equals("/IF")
                                                ||commandstr.equals("/DEFINE")) {
          cmdStack.pop();
          if (cmdStack.empty()) 
            throw new EParseException(fileName+":"+line+"  closing '"+commandstr+"' without valid opening");
          Command current=cmdStack.peek();
          if (current instanceof ForeachCmd) newcommands=((ForeachCmd)current).loopBody;
          else if (current instanceof FunctionDefinitionCmd) 
            newcommands=((FunctionDefinitionCmd)current).body;
          else if (current instanceof IfElseCmd) 
            if (((IfElseCmd)current)._elseBodyRead) newcommands=((IfElseCmd)current).elseBody;
            else newcommands=((IfElseCmd)current).ifBody;
        } else if (commandstr.equals("ELSE")) {
          newcommands=((IfElseCmd)cmdStack.peek()).elseBody; 
        } else if (functions.containsKey(commandstr)) {
          command=new FunctionCmd(functions.get(commandstr));
        }
        else throw new EParseException("unknown command "+commandstr+"\n in "+fileName+":"+line);
        
        if (command!=null) {
          for (String str: parts) {
            if (!"".equals(str))
              command.params.add(new Parameter(str));
          }
          command.file=fileName;
          command.line=line;
          commands.add(command);
        }
        if (newcommands!=null) {
          commands=newcommands;
          newcommands=null;
        }
        if (buffer.read()=='*') {
          //System.out.println("ignore ws");
          ignoreWhiteSpace=true;
          ignoreAllWhiteSpace=buffer.read()=='*';
          if (ignoreAllWhiteSpace) buffer.read();
        }
      }
    }
    if (read.length()>0)
      commands.add(new WriteCmd(new Parameter(read.toString(),false)));
    //System.out.print(read.toString());
    if (cmdStack.size()>1)
      throw new EParseException("reach file end of "+fileName+", but there are "+(cmdStack.size()-1)+" open blocks.");
    if (VERBOSE) if (fileName!=null && fileName!="") System.out.println("Ended parsing file \""+fileName+"\"");
    buffer.close();
    //if (transformationProperties!=null) transformationProperties.store(new FileOutputStream(parsedFile.getName()+".info"),"");
    return result;
  }
  
 
  //most recent  modified datetime of every loaded file
  /*long getNormalLastModified(){
    long lastModified=baseLastModified;
    for (File f: myLoadedFiles.values()) {
      long lm=f.lastModified();
      if (lm>lastModified)
        lastModified=lm;
    }  
    return lastModified;
  }*/
  
  
  /*String var_outputFile_extractedFileName(){
    if (createdFile==null) return "";
    return createdFile.getName();
  }*/
  

  

  /*String parseVariableValue()throws IOException{
    int ch;
    while  ((ch=buffer.read())==' ');
    StringBuffer v=new StringBuffer();
    while ((ch!=' ') && (ch!='}') && (ch!='"')&&(ch!=-1)){
      v.append((char)ch);
      ch=buffer.read();
    }
    if (ch=='}') _readbracket=true;
    return readVariable(v.toString());
  }
  
  String parseNext() throws IOException{
    int ch;
    while  ((ch=buffer.read())==' ');
    StringBuffer v=new StringBuffer();
    while ((ch!=' ') &&(ch!='}')&&(ch!=-1)){
      v.append((char)ch);
      ch=buffer.read();
    }
    if (ch=='}') _readbracket=true;
    return v.toString();
  }  
  */
  public static void printCommandHelp(){
    System.out.println("java FileParser <options> <file name>");
    System.out.println("Possible options:");
    System.out.println("\t-v\t\tVerbose mode");
    System.out.println("\t--regen\t\tWrites all output files, even if the input files have not changed");
    System.out.println("\t--print-changed-output-files\t\tPrints the filenames of modified output files");
    System.out.println("\t--print-result\t\tPrints the result of the processed main input file");    
  }
  
  public static void main(String[] arg) throws Exception{
    if (arg.length == 0) {
      printCommandHelp();
      return;
    }
    FileParser fp=new FileParser();
    String baseFile=arg[arg.length-1];
    try{
      fp.oldFileDates.load(new FileInputStream(baseFile+".info"));
    } catch (FileNotFoundException e){
      
    }
    boolean printResult = false;
    for (int i=0;i<arg.length-1;i++) {
      if (arg[i].equals("-v")) fp.VERBOSE = true;
      else if (arg[i].equals("-regen") || arg[i].equals("--regen")) fp.codeLastModified=9999999999999L;
      else if (arg[i].equals("--print-changed-output-files")) fp.printChangedOutputFiles = true;
      else if (arg[i].equals("--print-result")) printResult = true;
    }
    String res=fp.parseFile(baseFile).body.performAll();
    fp.fileDates.store(new FileOutputStream(baseFile+".info"),"");
    if (printResult) System.out.println(res);
  }
}
